@@ -1,3 +1,10 @@
+# To do:
+# Hammering prevention
+#  - Peer finding
+#  - Request trading
+# Negative cache
+# Max-age cache
+
 from twisted.python.failure import Failure
 
 from uuid import uuid4
@@ -23,12 +30,13 @@ from twisted.web import static
 import re
 import os
 
-from aws import sdb_now, sdb_now_plus
+from aws import sdb_now, sdb_now_add
 
 from resources.template import TemplateResource
 from resources.exposed import ExposedResource
 from resources.data import DataResource
 from resources.control import ControlResource
+from resources.peer import PeerResource
 
 from twisted.internet import threads
 
@@ -42,6 +50,9 @@ PrettyPrinter = pprint.PrettyPrinter(indent=4)
 import logging
 import logging.handlers
 
+from timeoffset import getTimeOffset
+from networkaddress import getNetworkAddress
+
 logger = logging.getLogger("main")
 
 def evaluateBoolean( b ):
@@ -69,40 +80,79 @@ class AWSpider:
     plugins = []
     start_time = time.time()
     
-    def __init__(self, aws_access_key_id, aws_secret_access_key, aws_s3_cache_bucket, aws_sdb_domain, aws_s3_storage_bucket=None, port=8080, max_requests=50, log_directory=None, log_level="debug" ):
+    public_ip = None
+    local_ip = None
+    network_information = {}
+    
+    uuid = uuid4().hex
+    reserved_arguments = ["reservation_function_name", "reservation_created", "reservation_next_request", "reservation_error"]
+    exposed_function_resources = {}
+    functions = {}
+    exposed_functions = []
+    peers = {}
+    peer_uuids = []
+    
+    def __init__( self, 
+                aws_access_key_id, 
+                aws_secret_access_key, 
+                aws_s3_cache_bucket, 
+                aws_sdb_reservation_domain, 
+                aws_s3_storage_bucket=None,
+                aws_sdb_coordination_domain=None, 
+                port=8080, 
+                max_requests=50, 
+                log_directory=None, 
+                log_level="debug",
+                web_admin=True,
+                web_admin_port=None,
+                time_offset=None,
+                peer_check_interval=5,
+                reservation_check_interval=5):
+        
+        self.time_offset = time_offset
+        self.peer_check_interval = int(peer_check_interval)
+        self.reservation_check_interval = int(reservation_check_interval)
         
         self.port = int(port)
+        self.network_information["port"] = self.port
+        
         self.rq = RequestQueuer( max_requests=int(max_requests) )
         
         self.aws_access_key_id = aws_access_key_id
         self.aws_secret_access_key = aws_secret_access_key
         self.aws_s3_cache_bucket = aws_s3_cache_bucket
         self.aws_s3_storage_bucket = aws_s3_storage_bucket
-        self.aws_sdb_domain = aws_sdb_domain
-            
-        self.exposed_function_resources = {}
-        
-        self.functions = {}
-        self.exposed_functions = []
+        self.aws_sdb_reservation_domain = aws_sdb_reservation_domain
+        self.aws_sdb_coordination_domain = aws_sdb_coordination_domain
         
         self.resource = Resource()
-        self.resource.putChild('', TemplateResource("index") )
-        self.resource.putChild('docs', TemplateResource() )
-        self.resource.putChild('static', static.File( os.path.join(os.path.dirname(__file__), 'web_static') ) )
+        
+        if web_admin_port is None:
+            web_admin_port = self.port
+        else:
+            web_admin_port = int(web_admin_port)
+
+        if web_admin:
+            self.resource.putChild('', TemplateResource("index" ) )
+            self.resource.putChild('static', static.File( os.path.join(os.path.dirname(__file__), 'web_static') ) )
+            self.resource.putChild('control', ControlResource( self ) )
+          
         self.resource.putChild('data', DataResource( self ) )
-        self.resource.putChild('control', ControlResource( self ) )
+        self.resource.putChild('peer', PeerResource( self ) )
         self.function_resource = Resource()
         self.resource.putChild( "function", self.function_resource )
         self.site = server.Site( self.resource )
         reactor.listenTCP(self.port, self.site)
 
+        if web_admin and web_admin_port != self.port:
+            self.web_admin_site = server.Site( self.resource )
+            reactor.listenTCP( web_admin_port, self.web_admin_site )
+
         self.s3 = AmazonS3( self.aws_access_key_id, self.aws_secret_access_key, self.rq )
         self.sdb = AmazonSDB( self.aws_access_key_id, self.aws_secret_access_key, self.rq )
         
         self.pg = PageGetter( self.s3, self.aws_s3_cache_bucket, rq=self.rq )
 
-        self.reserved_arguments = ["reservation_function_name", "reservation_created", "reservation_next_request", "reservation_error"]
-        
         if log_directory is None:
             handler = logging.StreamHandler()
         else:
@@ -124,11 +174,11 @@ class AWSpider:
             logger.setLevel(log_levels[log_level])
         else:
             logger.setLevel(logging.DEBUG)
-        
+    
         logger.info("Successfully loaded the Spider's configuration.")
 
     def showReservation( self, uuid ):
-        d = self.sdb.getAttributes( self.aws_sdb_domain, uuid )
+        d = self.sdb.getAttributes( self.aws_sdb_reservation_domain, uuid )
         return d
 
     def getServerData( self ):    
@@ -142,7 +192,7 @@ class AWSpider:
             "cost":cost
         }
         
-        logger.info("Got server data:\n%s" % PrettyPrinter.pformat(data) )
+        logger.debug("Got server data:\n%s" % PrettyPrinter.pformat(data) )
 
         return data
 
@@ -192,7 +242,7 @@ class AWSpider:
         if function["interval"] > 0:
             reserved_arguments = {}
             reserved_arguments["reservation_function_name"] = function_name
-            reserved_arguments["reservation_created"] = sdb_now()
+            reserved_arguments["reservation_created"] = sdb_now(offset=self.time_offset)
             reserved_arguments["reservation_next_request"] = reserved_arguments["reservation_created"]
             reserved_arguments["reservation_error"] = "0"
             
@@ -203,7 +253,7 @@ class AWSpider:
             uuid = uuid4().hex
             logger.debug( "Creating reservation on SimpleDB for %s, %s." % (function_name, uuid))
 
-            a = self.sdb.putAttributes( self.aws_sdb_domain, uuid, all_arguments )
+            a = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, all_arguments )
             a.addCallback( self._createReservationCallback, function_name, uuid )
             a.addErrback( self._createReservationErrback, function_name, uuid )
                         
@@ -349,80 +399,130 @@ class AWSpider:
         reactor.callWhenRunning( self._start )
         
     def _start( self ):
-        logger.critical( "Checking S3 and SimpleDB setup." )
+        logger.critical( "Checking S3 and SimpleDB setup, getting time offset for sync." )
         
         deferreds = []
         
         # Make sure the specified S3 cache bucket works.
-        a = self.s3.getBucket( self.aws_s3_cache_bucket )
-        a.addErrback( self._getS3BucketErrback, self.aws_s3_cache_bucket )
-        deferreds.append( a )
         
-        if self.aws_s3_storage_bucket is not None:
-            # Make sure the specified S3 storage bucket works.
-            b = self.s3.getBucket( self.aws_s3_storage_bucket )
-            b.addErrback( self._getS3BucketErrback, self.aws_s3_storage_bucket )
-            deferreds.append( b )
+        deferreds.append( self._checkAndCreateS3Bucket( self.aws_s3_cache_bucket ) )
 
-        # Make sure the specified SDB domain works.
-        c = self.sdb.domainMetadata( self.aws_sdb_domain )
-        c.addErrback( self._domainMetadataErrback )     
-        deferreds.append( c ) 
+        if self.aws_s3_storage_bucket is not None:
+            deferreds.append( self._checkAndCreateS3Bucket( self.aws_s3_storage_bucket ) )
+
+        deferreds.append( self._checkAndCreateSDBDomain( self.aws_sdb_reservation_domain ) )    
+
+        if self.aws_sdb_coordination_domain is not None:
+            deferreds.append( self._checkAndCreateSDBDomain( self.aws_sdb_coordination_domain ) )
+            deferreds.append( self.getNetworkAddress() )
+        
+        if self.time_offset is None:
+            deferreds.append( self.getTimeOffset() )
         
         d = DeferredList(deferreds, consumeErrors=True )
         d.addCallback( self._startCallback )
-        
-    def _startCallback( self, data ):
-        logger.critical( "Starting Spider." )
+
+    def _startCallback( self, data ):   
+        reactor.addSystemEventTrigger('before', 'shutdown', self._cleanupBeforeReactorShutdown)     
+        logger.critical( "Starting Spider UUID: %s" % self.uuid)
         self.queryloop = task.LoopingCall( self.query )
-        self.queryloop.start(5)
+        self.queryloop.start( self.reservation_check_interval )
+        if self.aws_sdb_coordination_domain is not None:
+            self.coordinateloop = task.LoopingCall( self.coordinate )
+            self.coordinateloop.start( self.peer_check_interval )  
+            reactor.callLater( 10, self.peerCheckRequest )          
         #self.paused = True
         pass
-    
-    # Bucket does not work.
-    def _getS3BucketErrback( self, error, bucket_name ):
+   
+    def getNetworkAddress( self ):
+        d = getNetworkAddress()
+        d.addCallback( self._getNetworkAddressCallback )
+        d.addErrback( self._getNetworkAddressErrback )
+        return d
+                
+    def _getNetworkAddressCallback( self, data  ):   
+        if "public_ip" in data:
+            self.public_ip = data["public_ip"]
+            self.network_information["public_ip"] = self.public_ip
+            
+        if "local_ip" in data:
+            self.local_ip = data["local_ip"]  
+            self.network_information["local_ip"] = self.local_ip
+         
+    def _getNetworkAddressErrback( self, error ):
+        logger.critical( "Could not get network address." )
+        self.shutdown()
+            
+    def getTimeOffset( self ):
+        d = getTimeOffset()
+        d.addCallback( self._getTimeOffsetCallback )
+        d.addErrback( self._getTimeOffsetErrback )
+        return d
         
+    def _getTimeOffsetCallback( self, time_offset ):
+        logger.info( "Got time offset for sync." )
+        self.time_offset = time_offset
+    
+    def _getTimeOffsetErrback( self, error ):
+        if self.time_offset is None:
+            logger.critical( "Could not get time offset for sync." )
+            self.shutdown()
+    
+    def _checkAndCreateS3Bucket( self, bucket_name ):
+        d = self.s3.getBucket( bucket_name )
+        d.addErrback( self._checkAndCreateS3BucketErrback, bucket_name ) 
+        return d
+        
+    def _checkAndCreateS3BucketErrback( self, error, bucket_name ):
         if int(error.value.status) == 404:
             logger.info( "Bucket %s does not exist, creating." % bucket_name )     
             d = self.s3.putBucket( bucket_name )
-            d.addErrback( self._getS3BucketErrback2, bucket_name )
+            d.addErrback( self._checkAndCreateS3BucketErrback2, bucket_name )
             return d
             
         logger.critical( "Could not find or create S3 bucket '%s'." % bucket_name )
         self.shutdown()
         
-    def _getS3BucketErrback2( self, error, bucket_name ):
+    def _checkAndCreateS3BucketErrback2( self, error, bucket_name ):
         logger.critical( "Could not create S3 bucket '%s'" % bucket_name )
         self.shutdown()
-    
-    def _domainMetadataErrback( self, error ):
+
+    def _checkAndCreateSDBDomain( self, domain_name ):    
+        d = self.sdb.domainMetadata( domain_name )
+        d.addErrback( self._checkAndCreateSDBDomainErrback, domain_name )     
+        return d
+
+    def _checkAndCreateSDBDomainErrback( self, error, domain_name ):
         
         if int(error.value.status) == 400:  
-            logger.info( "Domain %s does not exist, creating." % self.aws_sdb_domain )
-            d = self.sdb.createDomain( self.aws_sdb_domain  )
-            d.addErrback( self._domainMetadataErrback2 )
+            logger.info( "Domain %s does not exist, creating." % domain_name )
+            d = self.sdb.createDomain( domain_name  )
+            d.addErrback( self._checkAndCreateSDBDomainErrback2, domain_name )
             return d
         
-        logger.critical( "Could not find or create SDB domain '%s'." % self.aws_sdb_domain )
+        logger.critical( "Could not find or create SDB domain '%s'." % domain_name )
         self.shutdown()
         
-    def _domainMetadataErrback2( self, error ):
-        logger.critical( "Could not create SDB domain '%s'" % self.aws_sdb_domain )
+    def _checkAndCreateSDBDomainErrback2( self, error, domain_name ):
+        logger.critical( "Could not create SDB domain '%s'" % domain_name )
         self.shutdown()
     
     def pause(self):
         logger.critical( "Pausing Spider." )
         self.paused = True
         self.queryloop.stop() 
+        self.coordinateloop.stop()
         
     def resume(self):
         logger.critical( "Resuming Spider." )
         self.paused = False
-        self.queryloop.start(5) 
+        self.queryloop.start( self.reservation_check_interval ) 
+        if self.aws_sdb_coordination_domain is not None:
+            self.coordinateloop.start( self.peer_check_interval)
     
     def setReservationError( self, uuid, function_name="Unknown" ):
         logger.error( "Setting error on reservation %s, %s." % (function_name, uuid) )
-        d = self.sdb.putAttributes( self.aws_sdb_domain, uuid, {"reservation_error":"1"}, replace=["reservation_error"] )
+        d = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, {"reservation_error":"1"}, replace=["reservation_error"] )
         d.addCallback( self._setReservationErrorCallback, uuid )
         d.addErrback( self._setReservationErrorErrback, uuid )
         return d
@@ -435,7 +535,7 @@ class AWSpider:
     
     def deleteFunctionReservations( self, function_name ):
         logger.info( "Deleting reservations for function %s." % (function_name) )
-        d = self.sdb.select( "SELECT itemName() FROM %s WHERE reservation_function_name='%s'" % ( self.aws_sdb_domain, function_name) )
+        d = self.sdb.select( "SELECT itemName() FROM %s WHERE reservation_function_name='%s'" % ( self.aws_sdb_reservation_domain, function_name) )
         d.addCallback( self._deleteFunctionReservationsCallback, function_name )
         d.addErrback( self._deleteFunctionReservationsErrback, function_name )
         return d
@@ -444,7 +544,7 @@ class AWSpider:
         logger.debug( "Found reservations for function %s: %s" % (function_name, data) )
         deferreds = []
         for uuid in data:
-            deferreds.append( self.sdb.delete(self.aws_sdb_domain, uuid) )
+            deferreds.append( self.sdb.delete(self.aws_sdb_reservation_domain, uuid) )
         d = DeferredList(deferreds, consumeErrors = True )
         d.addCallback( self._deleteFunctionReservationsCallback2, function_name )
         return d
@@ -464,7 +564,7 @@ class AWSpider:
         
     def deleteReservation( self, uuid, function_name="Unknown" ):
         logger.info( "Deleting reservation %s, %s." % (function_name, uuid) )
-        d = self.sdb.delete( self.aws_sdb_domain, uuid )
+        d = self.sdb.delete( self.aws_sdb_reservation_domain, uuid )
         d.addCallback( self._deleteReservationCallback, function_name, uuid )
         d.addErrback( self._deleteReservationErrback, function_name, uuid )
         return d
@@ -485,7 +585,7 @@ class AWSpider:
         else:
             reservation_function_names = ""
         
-        sql = "SELECT * FROM %s WHERE reservation_next_request < '%s' INTERSECTION reservation_error != '1' %s" % ( self.aws_sdb_domain, sdb_now(), reservation_function_names )
+        sql = "SELECT * FROM %s WHERE reservation_next_request < '%s' INTERSECTION reservation_error != '1' %s" % ( self.aws_sdb_reservation_domain, sdb_now(offset=self.time_offset), reservation_function_names )
         
         #logger.debug( "Querying SimpleDB, \"%s\"" % sql )
         
@@ -537,7 +637,7 @@ class AWSpider:
                 logger.error( "Could not find resource %s." % function_name )
                 continue
             
-            d = self.sdb.putAttributes( self.aws_sdb_domain, uuid, {"reservation_next_request":sdb_now_plus(exposed_function["interval"]) }, replace=["reservation_next_request"] )
+            d = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, {"reservation_next_request":sdb_now_add(exposed_function["interval"], offset=self.time_offset) }, replace=["reservation_next_request"] )
             d.addCallback( self._setNextRequestCallback, function_name, uuid)
             d.addErrback( self._setNextRequestErrback, function_name, uuid )
             
@@ -620,8 +720,155 @@ class AWSpider:
     def _exposedFunctionErrback2( self, error, function_name, uuid ):
         logger.error( "Could not put results of %s, %s on S3.\n%s" % (function_name, uuid, error) )
 
+    
+    def _cleanupBeforeReactorShutdown( self ):
+        logger.debug( "Cleaning up before reactor shutdown." )
+        
+        deferreds = []
+        
+        if self.aws_sdb_coordination_domain is not None:
+            d = self.sdb.delete( self.aws_sdb_coordination_domain, self.uuid )
+            d.addCallback( self._cleanupBeforeReactorShutdownCallback )
+            return d
+    
+    def _cleanupBeforeReactorShutdownCallback( self, data  ):
+        logger.debug( "Deleted uuid from %s" % self.aws_sdb_coordination_domain )
+        return self.peerCheckRequest()
+    
+    def peerCheckRequest( self ):
+
+        logger.debug( "Signaling peers." )
+
+        deferreds = []
+
+        for uuid in self.peers:
+            if uuid != self.uuid:
+                logger.debug( "Signaling %s to check peers." % self.peers[uuid]["uri"] )
+                d = self.getPage( self.peers[uuid]["uri"] + "/peer/check", cache= -1 )
+                d.addCallback( self._peerCheckRequestCallback, self.peers[uuid]["uri"] )
+                deferreds.append( d )
+
+        if len(deferreds) > 0:
+            return DeferredList(deferreds)
+
+    def _peerCheckRequestCallback( self, data, uri ):
+        logger.debug( "Got %s/peer/check." % uri )
+    
     def shutdown( self ):
         logger.critical( "Stopping server." )
         reactor.callLater( 1, reactor.stop )
+    
+    def coordinate( self ):
+        print "Peers: %s" % self.peers
+        attributes = { "created":sdb_now(offset=self.time_offset) }
+        attributes.update( self.network_information )
+        self.sdb.putAttributes( self.aws_sdb_coordination_domain, self.uuid, attributes, replace=attributes.keys() )
+        reactor.callLater( 5, self.checkPeers )
+    
+    def checkPeers( self ):
+        sql = "SELECT public_ip, local_ip, port FROM %s WHERE created > '%s'" % ( self.aws_sdb_coordination_domain, sdb_now_add( self.peer_check_interval * -2, offset=self.time_offset) )
+        #logger.debug( "Querying SimpleDB, \"%s\"" % sql )
+        d = self.sdb.select( sql )
+        d.addCallback( self._checkPeersCallback )
+        d.addErrback( self._checkPeersErrback )
+        
+    def _checkPeersCallback( self, discovered ):
 
+        existing_peers = set( self.peers.keys() )
+        discovered_peers = set( discovered.keys() )
+        new_peers = discovered_peers - existing_peers
+        old_peers = existing_peers - discovered_peers
+        
+        deferreds = []
+        
+        for uuid in new_peers:
+            if uuid == self.uuid:
+                self.peers[uuid] = {
+                    "local_ip":"127.0.0.1",
+                    "port":self.port
+                }
+            else:
+                deferreds.append( self.verifyPeer( uuid, discovered[uuid] ) )
+        
+        for uuid in old_peers:
+            logger.debug( "Removing peer %s" % uuid )
+            del self.peers[ uuid ]
+        
+        if len( new_peers ) > 0:
+            if len(deferreds) > 0:
+                d = DeferredList(deferreds, consumeErrors=True)
+                d.addCallback( self._checkPeersCallback2 )
+            else:
+                self._checkPeersCallback2( None ) #Just found ourself.
+        elif len( old_peers ) > 0:
+            self._checkPeersCallback2( None )
+        else:
+            pass # No old, no new.
+        
+    def _checkPeersCallback2( self, data ):
+        logger.debug( "Re-organizing peers." )
+        for uuid in self.peers:
+            if "local_ip" in self.peers[uuid]:
+                self.peers[uuid]["uri"] = "http://%s:%s" % (self.peers[uuid]["local_ip"], self.peers[uuid]["port"] )
+            elif "public_ip" in self.peers[uuid]:
+                self.peers[uuid]["uri"] = "http://%s:%s" % (self.peers[uuid]["public_ip"], self.peers[uuid]["port"] )
+            else:
+                logger.error("Peer %s has no local or public IP. This should not happen." % uuid )
+        
+        self.peer_uuids = self.peers.keys()
+        self.peer_uuids.sort()
+        
+        print self.peers 
+        print self.peer_uuids
+        
+        # Go through the peers and reorganize the group.
+        pass
+        
+    def _checkPeersErrback( self, error ):
+        print error
 
+    def verifyPeer( self, uuid, peer ):
+        
+        logger.debug( "Verifying peer %s" % uuid )
+        
+        deferreds = []
+        
+        if "port" in peer:
+            port = int(peer["port"][0])
+        else:
+            port = self.port
+            
+        if "local_ip" in peer:
+            local_ip = peer["local_ip"][0]
+            local_url = "http://%s:%s/data/server" % (local_ip, port)
+            d = self.getPage( local_url, timeout=5, cache=-1 )
+            d.addCallback( self._verifyPeerLocalIPCallback, uuid, local_ip, port )
+            deferreds.append( d )
+            
+        if "public_ip" in peer:
+            public_ip = peer["public_ip"][0]
+            public_url = "http://%s:%s/data/server" % (public_ip, port)
+            d = self.getPage( public_url, timeout=5, cache=-1  )         
+            d.addCallback( self._verifyPeerPublicIPCallback, uuid, public_ip, port )
+            deferreds.append( d )
+
+        if len(deferreds) > 0:
+            d = DeferredList( deferreds, consumeErrors=True )
+            return d
+        else:
+            return None
+    
+    def _verifyPeerLocalIPCallback( self, data, uuid, local_ip, port ):
+        logger.debug( "Verified local IP for %s" % uuid )
+        if uuid not in self.peers:
+            self.peers[ uuid ] = {}
+        self.peers[ uuid ]["local_ip"] = local_ip
+        self.peers[ uuid ]["port"] = port
+    
+    def _verifyPeerPublicIPCallback( self, data, uuid, public_ip, port ):
+        logger.debug( "Verified public IP for %s" % uuid )
+        if uuid not in self.peers:
+            self.peers[ uuid ] = {}
+        self.peers[ uuid ]["public_ip"] = public_ip
+        self.peers[ uuid ]["port"] = port
+        
